/*
SimpleAlgebra.h
Written by Walter O. Krawec

Copyright (c) 2013 Walter O. Krawec

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef SIMPLE_ALGEBRA_H_
#define SIMPLE_ALGEBRA_H_

#include <math.h>
#include <stdlib.h>
#include <iostream>
#include <string>
#include <vector>

namespace algebra
{
	struct Complex16
	{
		Complex16() : real(0), imag(0) {};

		double real, imag;
	};

	class ComplexHelper
	{
	public:
		static Complex16 one() {Complex16 out; out.real = 1.0; out.imag = 0; return out;}
		static Complex16 zero() {Complex16 out; out.real = 0; out.imag = 0; return out;}

		static void sq(Complex16& v)
		{
			Complex16 v2;
			v2 = v;
			v2.imag *= -1;
			v = mult(v, v2);
		}

		static Complex16 construct(double theta)
		{
			Complex16 out;
			out.real = cos(theta);
			out.imag = sin(theta);
			return out;
		}

		static Complex16 power(Complex16& base, int e)
		{
			Complex16 a;
			a.real = log( sqrt( base.real*base.real + base.imag*base.imag ) );
			a.imag = atan2(base.imag, base.real);

			a.real *= e;
			a.imag *= e;

			Complex16 result = construct(a.imag);
			result.imag *= exp(a.real);
			result.real *= exp(a.real);

			return result;
		}

		static Complex16 add(Complex16& a, Complex16& b)
		{
			Complex16 out;
			out.real = a.real + b.real;
			out.imag = a.imag + b.imag;

			return out;
		}
		static Complex16 subtract(Complex16& a, Complex16& b)
		{
			Complex16 out;
			out.real = a.real - b.real;
			out.imag = a.imag - b.imag;

			return out;
		}

		static Complex16 mult(Complex16& a, Complex16& b)
		{
			Complex16 out;
			out.real = (a.real*b.real - a.imag*b.imag);
			out.imag = (a.imag*b.real + a.real*b.imag);

			return out;
		}
	};

	class mat
	{
	public:
		mat()
		{
			elements = NULL;
			r = 0;
			c = 0;
			n = 0;
		}
		mat(const mat& m)
		{
			elements = NULL;
			r = m.r;
			c = m.c;
			n = r*c;

			if(r == 0 || c == 0 || !m.elements)
				return;

			elements = new Complex16[n];

			for(int i=0; i<n; ++i)
				elements[i] = m.elements[i];
		}
		~mat() {cleanup();}

		mat& operator=(const mat& m)
		{
			if(!elements || r != m.r || c != m.c)
			{
				cleanup();
				r = m.r;
				c = m.c;
				n = r*c;

				if(r == 0 || c == 0 || !m.elements)
				{
					throw std::string("Error 1 in operator=\n");
					return *this;
				}

				elements = new Complex16[n];
			}

			if(!m.elements)
			{
				throw std::string("Error 2 in operator=\n");
				return *this;
			}

			for(int i=0; i<n; ++i)
				elements[i] = m.elements[i];

			return *this;
		}

		void cleanup()
		{
			if(elements)
				delete[] elements;

			elements = NULL;
			r = 0;
			c = 0;
			n = 0;
		}

		void zero()
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
				elements[i].imag = 0;
				elements[i].real = 0;
			}
		}

		void clear(Complex16 value)
		{
			if(!elements)
				return;
			for(int i=0; i<n; ++i)
				elements[i] = value;
		}

		bool create(int _r, int _c)
		{
			if(_r <= 0 || _c <= 0)
				return false;
			if(elements && r == _r && c == _c)
				return true;

			cleanup();
			r = _r;
			c = _c;
			n = r*c;

			elements = new Complex16[r*c];

			return true;
		}

		bool create(int _r)
		{
			return create(_r, _r);
		}

		Complex16& operator() (int row, int col)
		{
			int _r = row;
			int _c = col;
			if(!elements || _r >= r || _c >= c || _r < 0 || _c < 0)
			{
				std::cout << "Error 5 (" << elements << ", " << _r << ", " << r << ", " << _c << ", " << c << ")\n";
				throw(std::string("Error 1 in operator()"));
			}

			return elements[_c*r + _r];
		}
		Complex16& getDirect(int row, int col)
		{
			return (*this)(row, col);
		}

		void setDirect(int row, int col, Complex16 value)
		{
			(*this)(row,col) = value;
		}

		Complex16* getElements() {return elements;}

		void print(std::ostream& f, bool printAll=false)
		{
			if(!elements)
				return;
			int rStart = 0, cStart=0;
			int rows = r;
			int cols = c;
			if(rows > 10)
				rows = 10;
			if(cols > 10)
				cols = 10;

			if(printAll)
			{
				rows = r;
				cols = c;
			}
			for(int _r=rStart; _r < rows; ++_r)
			{
				for(int _c=cStart; _c<cols; ++_c)
					f << (*this)(_r,_c).real << " + " << (*this)(_r, _c).imag << "i\t";
				f << "\n";
			}

			f << "\n";
		}

		void multiplyScalar(Complex16& c)
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
				elements[i] = ComplexHelper::mult(elements[i], c);
			}
		}

		void ident()
		{
			if(!elements)
				return;

			zero();
			int dim = getRows();
			if(getCols() < getRows())
				dim = getCols();

			Complex16 one;
			one.real = 1.0;
			one.imag = 0.0;

			for(int i=0; i<dim; ++i)
				this->setDirect(i,i,one);
		}

		void subtract(Complex16& c)
		{
			if(!elements)
				return;

			for(int i=0; i<n; ++i)
			{
				elements[i].real = elements[i].real - c.real;
				elements[i].imag = elements[i].imag - c.imag;
			}
		}

		void transpose(mat& B)
		{
			if(!elements)
				return;
			B.create(c,r);
			for(int i=0; i<r; ++i)
			{
				for(int j=0; j<c; ++j)
				{
					B(j,i) = (*this)(i,j);
					B(j,i).imag *= -1;
				}
			}
		}

		void subtract(mat* B, mat* output)
		{
			if(!B || !output)
				return;

			if(!elements || !B->getElements() || getRows() != B->getRows() || getCols() != B->getCols())
			{
				throw std::string("Error - cannot subtract matrix\n");
				return;
			}

			output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
				output->elements[i].real = elements[i].real - B->getElements()[i].real;
				output->elements[i].imag = elements[i].imag - B->getElements()[i].imag;
			}
		}

		void add(mat* B, mat* output)
		{
			if(!B || !output)
				return;

			if(!elements || !B->getElements() || getRows() != B->getRows() || getCols() != B->getCols())
			{
				throw std::string("Error - cannot add matrix\n");
				return;
			}

			output->create(getRows(), getCols());

			for(int i=0; i<n; ++i)
			{
				output->elements[i].real = elements[i].real + B->getElements()[i].real;
				output->elements[i].imag = elements[i].imag + B->getElements()[i].imag;
			}
		}

		void dot(mat* B, Complex16& output);	// returns this <dot> B
		bool multiply(mat& B, mat& output);

		void tensor(mat* B, mat* output);

		void random_unitary(const std::vector <double>& phi, const std::vector <double>& psi, const std::vector <double>& chi);

		int getRows()
		{
			return r;
		}
		int getCols()
		{
			return c;
		}
		int getSize()
		{
			return n;
		}

	private:
		Complex16* elements;
		int r, c, n;
	};
};

typedef algebra::Complex16 Complex;

#endif